/****************************************************************************
 *
 * driver/dev.c
 */

/*********************************************
 * Device management support
 *********************************************/

#include <linux/version.h>
#if 0
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	#include <linux/config.h>
#else
	#include <linux/autoconf.h>
#endif
#endif
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/socket.h>
#include <linux/in.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/string.h>
#include <linux/sockios.h>
#include <linux/net.h>
#include <linux/fcntl.h>
#include <linux/mm.h>
#include <linux/interrupt.h>
#include <linux/proc_fs.h>
#include <linux/stat.h>
#include <linux/init.h>
#include <linux/poll.h>
#include <linux/delay.h>
#include <linux/netfilter_ipv4.h>
#include <linux/spi/spi.h>
#include <linux/gpio.h>
#include <linux/irq.h>

#include <linux/tcp.h>
#include <net/tcp.h>

#include <asm/uaccess.h>
#include <asm/unaligned.h>
#include <asm/system.h>
#include <linux/spinlock.h>

#include "w5500.h"

//#define DEBUG_PKT
#if 0
#define DPRINTK(format,args...) printk(format,##args)
#else
#define DPRINTK(format,args...)
#endif

#define LOCAL_PORT 	3010
#define power_on    1
#define power_down  0

/* 
 * Information that need to be kept for each board.
 */
struct net_local {	
	/* socket */
	struct sock	*gSocket[MAX_SOCK_NUM];
	
	/* network interface */
	struct net_device *dev;
	
	/* Rx buffer pointer */
	unsigned long	gCRxRdPtr[MAX_SOCK_NUM];
	unsigned long	gCRxWrPtr[MAX_SOCK_NUM];
	unsigned long	gCRxMinPtr[MAX_SOCK_NUM];
	unsigned long	gCRxMaxPtr[MAX_SOCK_NUM];

	/* Tx buffer pointer */
	unsigned long	gCTxRdPtr[MAX_SOCK_NUM];
	unsigned long	gCTxWrPtr[MAX_SOCK_NUM];
	unsigned long	gCTxFreePtr[MAX_SOCK_NUM];
	unsigned long	gCTxMinPtr[MAX_SOCK_NUM];
	unsigned long	gCTxMaxPtr[MAX_SOCK_NUM];

	/* status register */
	unsigned long	gStatus[MAX_SOCK_NUM];

	/* Tx control lock.  This protects the transmit buffer ring
	 * state along with the "tx full" state of the driver.  This
	 * means all netif_queue flow control actions are protected
	 * by this lock as well.
	 */
	spinlock_t		lock;
};


static int ENABLE_3V3_GPIO;
static int RESET_GPIO;
static int INT_GPIO;


/* prototypes */
static inline u8 iinchip_inb(u32 adr);
static inline void iinchip_outb(u32 adr, u8 val);
static void iinchip_mac_rx_tasklet(unsigned long data);
int iinchip_reset(void);

/* variables */
struct workqueue_struct *w5500_wq;

static spinlock_t spi_lock;
unsigned int local_port;
unsigned int SMASK[MAX_SOCK_NUM];
unsigned int RMASK[MAX_SOCK_NUM];
unsigned int SSIZE[MAX_SOCK_NUM];
unsigned int RSIZE[MAX_SOCK_NUM];
unsigned char *SBUFBASEADDRESS[MAX_SOCK_NUM];
unsigned char *RBUFBASEADDRESS[MAX_SOCK_NUM];
static struct net_local iinchip_local_p;
static unsigned char txsize[MAX_SOCK_NUM] = {
	16, 0, 0, 0, 0, 0, 0, 0
};
static unsigned char rxsize[MAX_SOCK_NUM] = {
	16, 0, 0, 0, 0, 0, 0, 0
};

extern struct spi_device *spi_device;
extern void w5500_tx_irq_work(struct work_struct *work);

/* macros */

/*
 * Print a hex-dump of the given block data
 */
void
iinchip_print_hexbuf(char *txt, char *data, int len)
{
	int i,j;

	printk("\n    ---- <%s:%d> ----\n", txt, len);
	for(i=0; i<len; i+=16) {
		printk("    %04X : ", i);
		for(j=i; (j<i+16) && (j<len);  j++) {
			printk("%02X ", (unsigned char)*(data+j));
		}
		printk("\n");
	}
	printk("    ----- <%s> ----\n\n", txt);
}

/*
 * Interrupt (un)lock.
 */
static inline void iinchip_irq_lock(void)
{
	unsigned int irq = gpio_to_irq(INT_GPIO);
	if (!in_irq()) disable_irq(irq);
}

static inline void iinchip_irq_unlock(void)
{
	unsigned int irq = gpio_to_irq(INT_GPIO);
	if (!in_irq()) enable_irq(irq);
}

void w5500_write(u32 addrbsb, u8 data)
{
	u8  addr[4];
	//unsigned long irqflags;

	addr[0] = (addrbsb & 0x00FF0000) >> 16;
	addr[1] = (addrbsb & 0x0000FF00) >> 8;
	addr[2] = (addrbsb & 0x000000F8) + 4;
	addr[3] = data;

	iinchip_irq_lock();
	//spin_lock_irqsave(&spi_lock, irqflags);

	spi_write(spi_device, addr, 4);

	//spin_unlock_irqrestore(&spi_lock, irqflags);
	iinchip_irq_unlock();
}

u8 w5500_read(u32 addrbsb)
{
	u8 data, addr[3];
	//unsigned long irqflags;

	addr[0] = (addrbsb & 0x00FF0000) >> 16;
	addr[1] = (addrbsb & 0x0000FF00) >> 8;
	addr[2] = (addrbsb & 0x000000F8);


	iinchip_irq_lock();
	//spin_lock_irqsave(&spi_lock, irqflags);

	spi_write_then_read(spi_device, addr, 3, &data, 1);	
//	printk("%s:data=%x\n",__func__,data);

	//spin_unlock_irqrestore(&spi_lock, irqflags);
	iinchip_irq_unlock();

	return data;
}

u16 w5500_write_buf(u32 addrbsb, const u8 *buf, u16 len)
{
	u8 addr[SPI_BURST_SIZE + 3];
	unsigned int i, l, r;
	//unsigned long irqflags;

	l = len / SPI_BURST_SIZE;
	r = len % SPI_BURST_SIZE;

	iinchip_irq_lock();
	//spin_lock_irqsave(&spi_lock, irqflags);

	for(i = 0 ; i < l ; i++)
	{
		addr[0] = (addrbsb & 0x00FF0000) >> 16;
		addr[1] = (addrbsb & 0x0000FF00) >> 8;
		addr[2] = (addrbsb & 0x000000F8) + 4;

		memcpy(addr + 3, buf + (i * SPI_BURST_SIZE), SPI_BURST_SIZE);
		spi_write_then_read(spi_device, addr, SPI_BURST_SIZE + 3, NULL, 0);
		addrbsb += (SPI_BURST_SIZE << 8);
	}

	if(r > 0)
	{
		addr[0] = (addrbsb & 0x00FF0000) >> 16;
		addr[1] = (addrbsb & 0x0000FF00) >> 8;
		addr[2] = (addrbsb & 0x000000F8) + 4;

		memcpy(addr + 3, buf + (i * SPI_BURST_SIZE), r);
		spi_write_then_read(spi_device, addr, r + 3, NULL, 0);
	}

	//spin_unlock_irqrestore(&spi_lock, irqflags);
	iinchip_irq_unlock();

	return len;
}

u16 w5500_read_buf(u32 addrbsb, u8 *buf, u16 len)
{
	u8 addr[3];
	unsigned int i, l, r;
	//unsigned long irqflags;

	l = len / SPI_BURST_SIZE;
	r = len % SPI_BURST_SIZE;

	iinchip_irq_lock();
	//spin_lock_irqsave(&spi_lock, irqflags);

	for(i = 0 ; i < l; i++)
	{
		addr[0] = (addrbsb & 0x00FF0000) >> 16 ;
		addr[1] = (addrbsb & 0x0000FF00) >> 8;
		addr[2] = (addrbsb & 0x000000F8);

		spi_write_then_read(spi_device, addr, 3, buf + (i * SPI_BURST_SIZE), SPI_BURST_SIZE);
		addrbsb += (SPI_BURST_SIZE << 8);
	}

	if(r > 0)
	{
		addr[0] = (addrbsb & 0x00FF0000) >> 16 ;
		addr[1] = (addrbsb & 0x0000FF00) >> 8;
		addr[2] = (addrbsb & 0x000000F8);
		spi_write_then_read(spi_device, addr, 3, buf + (i * SPI_BURST_SIZE), r);
	}

	//spin_unlock_irqrestore(&spi_lock, irqflags);
	iinchip_irq_unlock();

	return len;
}

static inline void iinchip_outb(u32 adr, u8 val)
{
	w5500_write(adr, val);
}

static inline void iinchip_outs(u32 adr, u16 val)
{
	char buf[2];

	buf[0] = ((val >> 8) & 0xff);
	buf[1] = (val  & 0xff);
	w5500_write_buf(adr, buf, 2);
}

static inline void iinchip_outl(u32 adr, u32 val)
{
	char buf[4];

	buf[0] = ((val >> 24) & 0xff);
	buf[1] = ((val >> 16) & 0xff);
	buf[2] = ((val >> 8) & 0xff);
	buf[3] = (val & 0xff);

	w5500_write_buf(adr, buf, 4);
}

static inline u8 iinchip_inb(u32 adr)
{
	return w5500_read(adr);
}

static inline u16 iinchip_ins(u32 adr)
{
	u16 val = 0;
	char buf[2];

	w5500_read_buf(adr, buf, 2);

	val = buf[0] << 8;
	val |= buf[1];
	
	return val;
}

static inline u32 iinchip_inl(u32 adr)
{
	u32 val = 0;
	char buf[4];

	w5500_read_buf(adr, buf, 4);
//	printk("%s(%d)=%x:%x:%x:%x:%x:%x\n", __func__,adr,buf[0], buf[1], buf[2], buf[3], buf[4], buf[5]);
	val = buf[0] << 24;
	val |= buf[1] << 16;
	val |= buf[2] << 8;
	val |= buf[3];

	return val;
}

void iinchip_copy_from(u32 adr, u8 *buf, ssize_t len)
{
	w5500_read_buf(adr, buf, len);
}

void iinchip_copy_to(u32 adr, const u8 *buf, ssize_t len)
{
	w5500_write_buf(adr, buf, len);
}

/*
 * Get Rx Buffer pointer.
 */
static void iinchip_get_rxbuffer(struct net_local *lp, int s)
{
	// get Rx Write pointer.
	lp->gCRxWrPtr[s] = iinchip_ins(RX_WR_PTR(s));
	// get Rx Read pointer.
	lp->gCRxRdPtr[s] = iinchip_ins(RX_RD_PTR(s));
	
	DPRINTK("s(%d): rx_rptr(0x%lx),rx_wptr(0x%lx)\n", s,
		lp->gCRxRdPtr[s], lp->gCRxWrPtr[s]);
}

/*
 * Get Rx Buffer Size.
 */
int iinchip_get_rxsize(int s)
{
	int l1, l2;
 retry:
	l1 = iinchip_ins(RX_RECV_SIZE_PTR(s));
	if (l1 == 0)
		return 0;
	l2 = iinchip_ins(RX_RECV_SIZE_PTR(s));
	if (l1 != l2)
		goto retry;

	DPRINTK("s(%d): rx_len(%d)\n", s, l1);
	
	return l1;
}

/*
 * Get Tx Buffer pointer.
 */
static void iinchip_get_txbuffer(struct net_local *lp, int s)
{
	// Get Tx Read pointer.
	lp->gCTxRdPtr[s] = iinchip_ins(TX_RD_PTR(s));
	// Get Tx Write pointer.
	lp->gCTxWrPtr[s] = iinchip_ins(TX_WR_PTR(s));
	
	DPRINTK("s(%d): tx_rptr(0x%lx),tx_wptr(0x%lx), tx_fptr(0x%lx)\n", s,
		lp->gCTxRdPtr[s], lp->gCTxWrPtr[s], lp->gCTxFreePtr[s] );
}

/*
 * Get Tx Buffer Size.
 */
int iinchip_get_txsize(int s)
{
	int l1, l2;
 retry:
	l1 = iinchip_ins(TX_FREE_SIZE_PTR(s));
	if (l1 == 0)
		return 0;
	l2 = iinchip_ins(TX_FREE_SIZE_PTR(s));
	if (l1 != l2)
		goto retry;
	
	return l1;
}

void iinchip_socket_setdport(int s, u16 port)
{
	iinchip_outs(DST_PORT_PTR(s), htons(port) );
}

void iinchip_socket_setdipaddr(int s, u32 addr)
{
	iinchip_outl(DST_IP_PTR(s), htonl(addr) );
}

void iinchip_socket_sethwaddr(char *addr)
{
	printk("%s(line=%d)=%x:%x:%x:%x:%x:%x\n", __FUNCTION__, __LINE__, 
			addr[0], addr[1], addr[2], addr[3], addr[4], addr[5]);
	iinchip_copy_to(REG_MAC_SRC, addr, 6);
}

void iinchip_setport(int s, int port)
{
	iinchip_outs(SRC_PORT_PTR(s), port);
}

#define MAX_CNT 100000 // udelay(2) * 100000 = 200msec. @2013.12.25 
#define _PHY_RESET_
//#define _CHIP_RESET_
int iinchip_socket_send(int s, int wait) 
{
	int free_buf_size;
    int cnt, ret;

	iinchip_outb(COMMAND(s), CSEND);
	while(iinchip_inb(COMMAND(s)));
    cnt = 0;
	ret = 1;

	do{ 
		udelay(2); 
		//udelay(2*10);
		
		free_buf_size = iinchip_get_txsize(s);
//		printk("%s:line=%d,buf-size=%d,s=%d,size=%d\n",__func__,__LINE__,free_buf_size,s,SSIZE[s]);
		cnt++;
		if(cnt>MAX_CNT)
		{
#ifdef _PHY_RESET_ 
			printk("\n----PHY RESET----\n");
			iinchip_outb(REG_PHYCFGR,0x38); //PHY RESET - Reset bit is low
			udelay(50); 
			iinchip_outb(REG_PHYCFGR,0xC8); //PHY RESET - Reset bit is high
#else //_CHIP_RESET
			printk("----W5500 RESET----\n");
			iinchip_reset();
 
			iinchip_close(s);
			mdelay(10);
			iinchip_open(s, WSOCK_MACL_RAWM, 0); 
#endif	
			cnt = 0;
			ret = 0;
			break;
		}
	}while(free_buf_size != SSIZE[s]);

	return ret;
}

void iinchip_socket_recv(int s)
{
	iinchip_outb(COMMAND(s), CRECV);
	while(iinchip_inb(COMMAND(s)));
}

void iinchip_rxrdptr_update(int s, unsigned long addr)
{
	iinchip_outs(RX_RD_PTR(s), addr);
}

void iinchip_txwrptr_update(int s, unsigned long addr)
{
	iinchip_outs(TX_WR_PTR(s), addr);
}

int iinchip_send_buf(int s, unsigned char *buf, int len)
{
	u32 addrbsb;
	u16 tx_ptr;
	int bufsize;
	u16 send_succ;

	DPRINTK("%s: s(%d), tx_len(%d),LINE=%d\n", __FUNCTION__, s, len,__LINE__);

	/* check buffer */
	bufsize = iinchip_get_txsize(s);
	DPRINTK("%s: s(%d), tx_len(%d),LINE=%d\n", __FUNCTION__, s, len,__LINE__);
	if (!bufsize || len > bufsize) {
		printk("s(%d): tx buffer(%d) is not enough for pkt(%d)\n", s, bufsize, len);
		return -1;
	}

	/* get write offset address */
	tx_ptr = iinchip_ins(TX_WR_PTR(s));
	addrbsb = (u32)(tx_ptr << 8) + (s << 5) + 0x10;
	DPRINTK("%s: s(%d), tx_len(%d),LINE=%d\n", __FUNCTION__, s, len,__LINE__);

	/* write to buffer */
	w5500_write_buf(addrbsb, buf, len);
	DPRINTK("%s: s(%d), tx_len(%d),LINE=%d\n", __FUNCTION__, s, len,__LINE__);

	/* update write offset address */
	tx_ptr += len;
	iinchip_txwrptr_update(s, tx_ptr);

	/* command send */
	send_succ = iinchip_socket_send(s,0);
    if(send_succ== 0)
		len = -1;

#ifdef DEBUG_PKT
	iinchip_print_hexbuf("Tx", buf, len);
#endif

	return len;
}

int iinchip_send_msg(int s, unsigned char *data, int len)
{
	int buf_len, snd_len, req_len;
	unsigned char *snd_ptr;
	u32 addrbsb;
	u16 tx_ptr;
	
	/* verify_area is done by net/socket.c */
	snd_ptr = data;
	req_len = len;
	DPRINTK("%s: s(%d),len(%d)\n", __FUNCTION__, s, req_len);
	
	while (req_len > 0) {
		
		buf_len = iinchip_get_txsize(s);
		if (buf_len == 0) {
			continue;
		}
		
		snd_len = (buf_len > req_len) ? req_len : buf_len;
		if (snd_len > 0) {
			/* get write offset address */
			tx_ptr = iinchip_ins(TX_WR_PTR(s));
			addrbsb = (u32)(tx_ptr << 8) + (s << 5) + 0x10;

			/* write to buffer */
			w5500_write_buf(addrbsb, snd_ptr, snd_len);

			/* update write offset address */
			tx_ptr += snd_len;
			iinchip_txwrptr_update(s, tx_ptr);

			/* command send */
			iinchip_socket_send(s, 1);

			/* W5500 ADDED */
			while((iinchip_inb(INT_STATUS(s)) & ISR_SEND_OK) != ISR_SEND_OK);
			iinchip_outb(INT_STATUS(s), ISR_SEND_OK);

			req_len -= snd_len;
			snd_ptr += snd_len;
			if (req_len <= 0)
				break;
		}
	}
	
	return len;
}

int iinchip_recv_msg(int s, unsigned char *buf, int buflen)
{
	int len, maxlen = RSIZE[s];
	u32 addrbsb;
	u16 rd_ptr;
	
	/* Get Rx Buffer Size */
	len = iinchip_get_rxsize(s);
	if (len == 0) {
		return 0;
	}

	if (len > maxlen) {
		printk("%s: length(%d) is invalid, drop frame.\n", DRV_NAME, len);
		return -1;
	}
	
	if (len > buflen)
		len = buflen;
	
	/* get read pointer */
	rd_ptr = iinchip_ins(RX_RD_PTR(s));
	addrbsb = (u32)(rd_ptr << 8) + (s << 5) + 0x18;
	
	/* read from buffer */
	w5500_read_buf(addrbsb, buf, len);
	
	/* set read pointer */
	rd_ptr += len;
	iinchip_outs(RX_RD_PTR(s), rd_ptr);
	
	/* command recv */
	iinchip_socket_recv(s);
	
	return len;
}

static void 
iinchip_read_rx_buf(struct net_local *lp, int s, u8 *buf, int len)
{
	u32 rptr, addrbsb;

	/* read from ASIC's buffer */
	rptr = lp->gCRxRdPtr[s];

	addrbsb = (u32)(rptr << 8) + (s << 5) + 0x18;

	/* read from buffer */
	w5500_read_buf(addrbsb, buf, len);

#ifdef DEBUG_PKT
	iinchip_print_hexbuf("Rx", buf, len);
#endif
}

static void 
iinchip_update_rx_bufptr(struct net_local *lp, int s, int len)
{
	/* update address */
	lp->gCRxRdPtr[s] += len;
}


/************************************************************************
 *
 * Wiznet Socket Management Routines.
 *
 ***********************************************************************/
int iinchip_netif_create(struct net_device *dev)
{
	struct net_local *lp = &iinchip_local_p;
	struct wiz_private *wp = netdev_priv(dev);

	if (lp->dev) {
		return -EBUSY;
	}

	lp->dev = dev;
	iinchip_socket_interrupt_enable(wp->s);
	
	return 0;
}

int iinchip_netif_delete(struct net_device *dev)
{
	struct net_local *lp = &iinchip_local_p;
	struct wiz_private *wp = netdev_priv(dev);
	
	if (lp->dev != dev) {
		return -EFAULT;
	}

	iinchip_socket_interrupt_disable(wp->s);
	lp->dev = NULL;
	
	return 0;
}

int iinchip_socket_interrupt_enable(int s)
{
	unsigned char mask;
	
	if (s < 0 || s >= MAX_SOCK_NUM)
		return -1;

	mask = iinchip_inb(REG_INT_MASK);
	mask |= (0x01 << s);
	iinchip_outb(REG_INT_MASK, mask);
	
	return 0;
}

int iinchip_socket_interrupt_disable(int s)
{
	unsigned char mask;

	if (s < 0 || s >= MAX_SOCK_NUM)
		return -1;

	mask = iinchip_inb(REG_INT_MASK);
	mask &= ~(0x01 << s);
	iinchip_outb(REG_INT_MASK, mask);
	
	return 0;
}

int iinchip_socket_status(int s)
{
	if (s < 0 || s >= MAX_SOCK_NUM)
		return -1;

	return iinchip_inb(SOCK_STATUS(s));
}

/************************************************************************
 *
 * Interrupt related routines.
 *
 ***********************************************************************/
static inline int
iinchip_socket_get_mtu(int type)
{
	if (type == SOCK_STREAM)
		//return 1450;
		return 1460;
	
	return 1480;
}

static void iinchip_mac_rx_tasklet(unsigned long data)
{
	struct net_local *lp = &iinchip_local_p;
	struct net_device *dev = (struct net_device *) data;
	struct wiz_private *wp = netdev_priv(dev);
	struct sk_buff *skb;
	int s, len, maxlen;
	DPRINTK("%s: start\n", __func__);
	
	s = wp->s;
	
	/* Get Rx Buffer Size */
	len = iinchip_get_rxsize(s);
	if (len == 0) {
		return;
	}
	
	maxlen = RSIZE[s];
	if (len > maxlen) {
		//DPRINTK
		printk("%s: length(%d) is invalid, reset socket.\n", dev->name, len);
		wp->stats.rx_errors++;
		iinchip_close(s);
		mdelay(10);
		iinchip_open(s, WSOCK_MACL_RAWM, 0);
		return;
	}
	
	/* Get Rx Buffer Pointer */
	iinchip_get_rxbuffer(lp, s);

	while (len > 0) {
		u8 t[4];

		int pktlen;

		iinchip_read_rx_buf(lp, s, t, 2);
		iinchip_update_rx_bufptr(lp, s, 2);
		
		pktlen = (t[0] << 8) + t[1] - 2;
		if(pktlen > 1514)
		{
			printk("Packet Size Error 1514>> %d\n", pktlen);
			wp->stats.rx_errors++;
			iinchip_close(s);
			mdelay(10);
			iinchip_open(s, WSOCK_MACL_RAWM, 0);
			return;
		}

#if 0	// 2014.02.27 sskim
		if (pktlen < 60) {
			printk("%s: Invalid pktlen(%d), dropping packet.\n", dev->name, pktlen);
			wp->stats.rx_dropped++;
			/* update buffer pointer */
			iinchip_update_rx_bufptr(lp, s, len-2);
			iinchip_rxrdptr_update(s, lp->gCRxRdPtr[s]);
			/* command recv */
			iinchip_socket_recv(s);
			return;
		}
#endif

		/* allocate socket buffer */
		skb = dev_alloc_skb(pktlen + 2);
		if (!skb) {
			printk("%s: Memory squeeze, dropping packet.\n", dev->name);
			wp->stats.rx_dropped++;
			/* update buffer pointer */
			iinchip_update_rx_bufptr(lp, s, len-2);				
			iinchip_rxrdptr_update(s, lp->gCRxRdPtr[s]);
			/* command recv */
			iinchip_socket_recv(s);
			return;
		}
		skb->dev = dev;
		skb_reserve(skb, 2);
		skb_put(skb, pktlen);
		
		/* read data */
		iinchip_read_rx_buf(lp, s, skb->data, pktlen);
		iinchip_update_rx_bufptr(lp, s, pktlen);
		
		/* Send the packet to the upper layers */
		skb->protocol = eth_type_trans(skb, dev);
		netif_rx(skb);

		/* update counter */
		wp->stats.rx_bytes += pktlen;
		wp->stats.rx_packets++;

		/* update buffer pointer */
		iinchip_rxrdptr_update(s, lp->gCRxRdPtr[s]);
		
		/* command recv */
		iinchip_socket_recv(s);

		len = iinchip_get_rxsize(s);
		if (len == 0) {
			return;
		}
	}
}

static void 
iinchip_irq_event(struct net_local *lp, int s)
{
	struct sock *sk = lp->gSocket[s];

	if (sk) {
		lp->gStatus[s] = iinchip_inb(SOCK_STATUS(s));
		DPRINTK("%s: SOCK Status(0x%lx)\n", DRV_NAME, lp->gStatus[s]);
		
		//if (lp->gStatus[s] & SOCK_ESTABLISHED) {
		if (lp->gStatus[s] == SOCK_ESTABLISHED) {
			sk->sk_state = TCP_ESTABLISHED;
			wake_up_interruptible(sk_sleep(sk));
		//} else if (lp->gStatus[s] & SOCK_CLOSED) {
		} else if (lp->gStatus == SOCK_CLOSED) {
			sk->sk_state = TCP_CLOSE;
			wake_up_interruptible(sk_sleep(sk));
		}
	}
}

static void w5500_rx_irq_work(struct work_struct *work)
{
	struct net_local *lp=&iinchip_local_p;
	int s;
	u8 isr, ssr;
	
	unsigned int irq = gpio_to_irq(INT_GPIO);
	DPRINTK("%s: start\n", __func__);
	
	isr = iinchip_inb(REG_SIR);

	DPRINTK("%s: channel_0=%d\n", __func__,iinchip_inb(OPT_PROTOCOL(0)));

	if (!isr) {
		iinchip_outb(REG_SIR, isr);
		enable_irq(irq);
		return;
	}
	
	/* clear global interrupt */
	iinchip_outb(REG_SIR, isr);
	
	/* socket interrupt */
	s = 0;
	if (isr & (1<<s)) {
		ssr = iinchip_inb(INT_STATUS(s));
		DPRINTK("%s: socket(%d) ssr(0x%x)\n", DRV_NAME, s, ssr);
		/* clear socket interrupt */
		
		if (s > 0 && (ssr & ISR_CON)) {
			DPRINTK("%s: socket(%d) ISR_CON\n", DRV_NAME, s);
			iinchip_irq_event(lp, s);
		}
		if (s > 0 && (ssr & ISR_DISCON)) {
			DPRINTK("%s: socket(%d) ISR_DISCON\n", DRV_NAME, s);
			iinchip_irq_event(lp, s);
		}
		if (s > 0 && (ssr & ISR_TIMEOUT)) {
			DPRINTK("%s: socket(%d) ISR_TIMEOUT\n", DRV_NAME, s);
			iinchip_irq_event(lp, s);
		}
		if (ssr & ISR_RECV) {
			iinchip_mac_rx_tasklet((unsigned long)lp->dev);
		}

		iinchip_outb(INT_STATUS(s), ssr&0x0f);
	}

	enable_irq(irq);
}

static irqreturn_t
iinchip_interrupt(int irq, void *dev_id)
{
	wiz_t *wz = (wiz_t *)dev_id;

	disable_irq_nosync(irq);
	queue_work(w5500_wq, &(wz->rx_work));
	return IRQ_HANDLED;
}

/************************************************************************
 *
 * Socket related routines.
 *
 ***********************************************************************/
int iinchip_open(int s, int type, int ipproto)
{
	struct net_local *lp=&iinchip_local_p;

	// Socket protocol.
	switch (type) {
	case WSOCK_MACL_RAWM:
		if (s != 0) {
			printk("%s: Not Supported\n", DRV_NAME);
			return -EFAULT;
		}

		iinchip_outb(OPT_PROTOCOL(s), WSOCKOPT_UCASTB | WSOCK_MACL_RAWM);
		iinchip_outb(REG_SIMR, 0x01);
		iinchip_outb(INT_MASK(0), 0x0f);
		break;
	default:
		printk("W5500: Unknown socket type (%d)\n", type);
		return -EFAULT;
	}

	// Initialize Rx Buffer Pointer.
	lp->gCRxMinPtr[s] = (uint32_t) RBUFBASEADDRESS[s];
	lp->gCRxMaxPtr[s] = (uint32_t) RBUFBASEADDRESS[s] + RSIZE[s];
	
	// Initialize Tx Buffer Pointer.
	lp->gCTxWrPtr[s] = (uint32_t) SBUFBASEADDRESS[s];
	lp->gCTxRdPtr[s] = (uint32_t) SBUFBASEADDRESS[s];
	lp->gCTxFreePtr[s] = (uint32_t) SBUFBASEADDRESS[s];
	lp->gCTxMinPtr[s] = (uint32_t) SBUFBASEADDRESS[s];
	lp->gCTxMaxPtr[s] = (uint32_t) SBUFBASEADDRESS[s] + SSIZE[s];
	
	// Send command.
	lp->gStatus[s] = 0;
	iinchip_outb(COMMAND(s), CSOCKINIT);
	while(iinchip_inb(COMMAND(s)) & CSOCKINIT);
	
	DPRINTK("sock:%d SOCK_STATUS = 0x%x , Protocol = 0x%x\n", s, iinchip_inb(SOCK_STATUS(s)), iinchip_inb(OPT_PROTOCOL(s)));	
	return 0;
}

int iinchip_close(int s)
{
	struct net_local *lp=&iinchip_local_p;
	unsigned long flags;
	
	if (s < 0 || s >= MAX_SOCK_NUM)
		return -1;

	local_irq_save(flags);			// for test sskim TODO
	iinchip_get_rxbuffer(lp, s);
	iinchip_get_txbuffer(lp, s);
	local_irq_restore(flags);		// for test sskim TODO

	/* Re-set Tx-pointer. */
	iinchip_txwrptr_update(s, lp->gCTxWrPtr[s]);

	// Send command.
	iinchip_outb(COMMAND(s), CCLOSE);
	while(iinchip_inb(COMMAND(s)) & CCLOSE);
	
	return 0;
}

/************************************************************************
 *	
 * Startup related routines.
 * 	
 ***********************************************************************/
void iinchip_sysinit(void)
{
	int i, ssum=0, rsum=0;

	for (i = 0 ; i < MAX_SOCK_NUM; i++) {

		/* Set Buffer Size */
		iinchip_outb(TX_BUF_SIZE_PTR(i), txsize[i]);
		iinchip_outb(RX_BUF_SIZE_PTR(i), rxsize[i]);

		SSIZE[i] = 0;
		RSIZE[i] = 0;

		if (ssum <= 16384) {
			switch (txsize[i]) {
			case 0:
				SSIZE[i] = 0;
				SMASK[i] = 0;
				break;
			case 2:	
				SSIZE[i] = 2048;
				SMASK[i] = 0x7FF;
				break;
			case 4:
				SSIZE[i] = 4096;
				SMASK[i] = 0xFFF;
				break;
			case 8:	
				SSIZE[i] = 8192;
				SMASK[i] = 0x1FFF;
				break;
			case 16:
				SSIZE[i] = 16384;
				SMASK[i] = 0x3FFF;
				break;
			default:
				SSIZE[i] = 2048;
				SMASK[i] = 0x7FF;
				break;
			}
		}
		if (rsum <= 16384) {
			switch (rxsize[i]) {
			case 0:
				SSIZE[i] = 0;
				SMASK[i] = 0;
				break;
			case 2:	
				RSIZE[i] = 2048;
				RMASK[i] = 0x7FF;
				break;
			case 4:	
				RSIZE[i] = 4096;
				RMASK[i] = 0xFFF;
				break;
			case 8:	
				RSIZE[i] = 8192;	
				RMASK[i] = 0x1FFF;
				break;
			case 16:
				RSIZE[i] = 16384;
				RMASK[i] = 0x3FFF;
				break;
			default:
				RSIZE[i] = 2048;
				RMASK[i] = 0x7FF;
				break;
			}
		}
		ssum += SSIZE[i];
		rsum += RSIZE[i];

		SBUFBASEADDRESS[i] = (unsigned char *) (SEND_DATA_BUF);
		RBUFBASEADDRESS[i] = (unsigned char *) (RECV_DATA_BUF);
		
		printk("%s: socket(%d) Tx(0x%x,0x%x,0x%x) Rx(0x%x,0x%x,0x%x)\n",
			DRV_NAME, i,
			SSIZE[i], SMASK[i], (uint32_t) SBUFBASEADDRESS[i],
			RSIZE[i], RMASK[i], (uint32_t) RBUFBASEADDRESS[i] );
	}
}

static u8 iinchip_macaddr[6];

/*
 */
void iinchip_hwreset(void)
{
	gpio_request(RESET_GPIO, "w5500_reset");
	gpio_direction_output(RESET_GPIO, 1);
	gpio_set_value(RESET_GPIO, 0);
	mdelay(1);
	gpio_set_value(RESET_GPIO, 1);
	mdelay(2);
	mdelay(1500);				// WIZ550IO 
}

void iinchip_power(u8 value)
{
	gpio_set_value(ENABLE_3V3_GPIO, value);	
	mdelay(1);				 
}

extern void iinchip_up_or_shutdown(u8 input)
{
	u8 value = input;
	if(value){
		iinchip_power(power_on);
		iinchip_reset();
	}else{
		iinchip_power(power_down);
	}		
}

int iinchip_reset(void)
{
	unsigned char addr[8];

	/* Initialize ASIC */

	iinchip_hwreset();

	printk("Version : %02x\n",iinchip_inb(REG_VERSIONR));
	iinchip_outb(REG_TMODE, TMODE_PINGBLOCK | TMODE_NOSIZECHK_RAW);

	// HW address.
	iinchip_socket_sethwaddr(iinchip_macaddr);
	iinchip_copy_from(REG_MAC_SRC, addr, 6);
	printk("%s: MAC [%02x:%02x:%02x:%02x:%02x:%02x]\n", DRV_NAME,
		addr[0],addr[1],addr[2],addr[3],addr[4],addr[5] );

	iinchip_sysinit();
	
	// Interrupt.
	iinchip_outb(REG_INT_MASK, 0x00);

	return 0;
}


/*
 *	Initialize the WIZnet TCP/IP protocol family
 */
int wiz_dev_init(wiz_t *wz)
{
	spin_lock_init(&spi_lock);
	printk("wiz_dev_init \n");

	w5500_wq = create_workqueue("w5500_workqueue");
	INIT_WORK(&(wz->rx_work), w5500_rx_irq_work);
	INIT_WORK(&(wz->tx_work), w5500_tx_irq_work);
	
	RESET_GPIO = wz->res_gpio;
	ENABLE_3V3_GPIO = wz->en_gpio;
	INT_GPIO = wz->int_gpio;
	gpio_request(ENABLE_3V3_GPIO, "w5500_3v3_enable");
	gpio_direction_output(ENABLE_3V3_GPIO, 1);
	iinchip_power(power_on);	
	/* make room for the local structure containing stats etc */
	memset(&iinchip_local_p, 0, sizeof(iinchip_local_p));
	
	/* initialize parameter */
	memcpy(iinchip_macaddr, wz->macaddr, 6);
	
	/* Initialize channel */
	local_port = LOCAL_PORT;

	/* Initialize ASIC */
	iinchip_reset();

	/* Initialize Interrupt GPIO */
	if(gpio_request(INT_GPIO, "w5500")) {
		printk("[%s] gpio_request w5500 irq error\n", __func__);
		return -EAGAIN;
	}	
	gpio_direction_input(INT_GPIO);

	/* allocate the irq corresponding to the receiving */
	if (request_irq(wz->irq, iinchip_interrupt, (IRQF_SHARED | IRQF_DISABLED | IRQF_TRIGGER_LOW), "w5500", (void *) wz)) {
		printk("errro request irq(%d)\n", wz->irq);		
		return -EAGAIN;
	}

	return 0;
}

int wiz_dev_exit(wiz_t *wz)
{
	destroy_workqueue(w5500_wq);
	free_irq(wz->irq, (void *) wz);

	printk("gpio_free\n");
	gpio_free(INT_GPIO);
	gpio_free(RESET_GPIO);

	return 0;
}
