/****************************************************************************
 *
 * driver/module.c
 */

/*********************************************
 * Module support
 *********************************************/

#include <linux/gpio.h>
#include <linux/of_gpio.h>
#include <linux/spi/spi.h>
#include "w5500.h"

#define test


static wiz_t gDrvInfo;

#ifdef test
static struct proc_dir_entry *w5500_proc_entry;
static int w5500_power_on_flag=0;
#endif

static unsigned char defmac[] = { 0x00, 0x08, 0xDC, 0x91, 0x97, 0x98 };

struct spi_device *spi_device = NULL;

#define W5500_PROC_NAME "w5500"

static int netdev_parse_dt(struct device *dev, struct netcard *pdata)
{
	struct device_node *np = dev->of_node;

	pdata->int_gpio = of_get_named_gpio(np, "int-gpio", 0);
	pdata->reset_gpio = of_get_named_gpio(np, "rst-gpio", 0);
	pdata->en_3p3_gpio = of_get_named_gpio(np, "3v3-en-gpio", 0);
	pr_debug("%s:int-gpio=%d,rst-gpio=%d,3v3-en-gpio=%d\n",
		__func__,pdata->int_gpio,pdata->reset_gpio,pdata->en_3p3_gpio);
	return 0;
}


#ifdef test
static int w5500_flag_show(struct seq_file *user_buf, void *v)
{
	seq_printf(user_buf,"%d\n",w5500_power_on_flag);
	return 0;
}

static s32 w5500_flag_write(struct file *file,const char __user *user_buf,
					size_t count, loff_t *ppos)
{
	int value;
	
	if (sscanf(user_buf, "%d", &value)!=1) {
		printk("%s:copy from user error\n", __func__);
		return -EFAULT;
	}
	printk("%s:buffer=%d,w5500_power_on_flag=%d\n", __func__,value,w5500_power_on_flag);
	if((1 == value)&&(1 == w5500_power_on_flag))
	{
	}else if((1 == value)&&(0 == w5500_power_on_flag)){
		iinchip_up_or_shutdown(true);		
		w5500_power_on_flag=1;
	}else if((0 == value)&&(0 == w5500_power_on_flag)){
	}else if((0 == value)&&(1 == w5500_power_on_flag)){
		iinchip_up_or_shutdown(false);		
		w5500_power_on_flag=0;
	}
	return count;
}

static int w5500_proc_open(struct inode *inode, struct file *file)
{
	return single_open(file, w5500_flag_show, NULL);
}

static const struct file_operations w5500_proc_fops = {
	.owner = THIS_MODULE,
	.write = w5500_flag_write,
	.read = seq_read,
	.open = w5500_proc_open,
	.llseek 	= seq_lseek,
	.release	= single_release,
};

#endif

static int w5500_probe(struct spi_device *spi)
{
	struct netcard *netdev;
	int ret;
	printk("w5500 probe\n");
	spi_device = spi;

	netdev = devm_kzalloc(&spi->dev, sizeof(*netdev),GFP_KERNEL);
	if(!netdev){
		pr_err("spi fail to allocate netdev\n");
		return -ENOMEM;
	}
	netdev->spi = spi;
	if(netdev->spi->dev.of_node){
		ret = netdev_parse_dt(&netdev->spi->dev,netdev);
		if(ret)
			return ret;
	}else{
		pr_err("netdev have not of_node\n");
		return -EIO;
	}
	
	/* Initial field */
	gDrvInfo.base = 0;
	gDrvInfo.irq  = gpio_to_irq(netdev->int_gpio);
	gDrvInfo.en_gpio = netdev->en_3p3_gpio;
	gDrvInfo.res_gpio = netdev->reset_gpio;
	gDrvInfo.int_gpio = netdev->int_gpio;

	/* mac address */
	memcpy(gDrvInfo.macaddr, defmac, 6);
	
	/* initialize device */
	if (wiz_dev_init(&gDrvInfo) < 0) {
		return -EFAULT;
	}

	/* create network interface */
	gDrvInfo.dev = wiznet_drv_create(&gDrvInfo);

#ifdef test
	w5500_proc_entry = proc_create(W5500_PROC_NAME,
			S_IWUSR | S_IWGRP | S_IRUSR | S_IRGRP,
			w5500_proc_entry,
			&w5500_proc_fops);
	if (w5500_proc_entry == NULL) {
		dev_err(&spi->dev, "Couldn't create proc entry!");
		return -1;
	}
#endif
	return 0;
}

static int w5500_remove(struct spi_device *spi)
{
	printk("w5500 remove\n");

	/* de-initialize device */
	if (gDrvInfo.dev) {
		wiznet_drv_delete(gDrvInfo.dev);
		gDrvInfo.dev = NULL;
	}

	/* free irq */
	wiz_dev_exit(&gDrvInfo);

	spi_device = NULL;
	
#ifdef test
	remove_proc_entry(W5500_PROC_NAME, NULL);
#endif

	return 0;
}
static struct of_device_id w5500_of_match[]={
	{.compatible = "qcom,w5500"},
	{},
};
static struct spi_driver w5500_driver = {
	.driver = {
		.name = "w5500",
		.bus = &spi_bus_type,
		.owner = THIS_MODULE,
		.of_match_table=w5500_of_match,
	},
	.probe = w5500_probe,
	.remove = __devexit_p(w5500_remove),
};

static int __init
wiz_module_init(void)
{
	int ret;

	printk("%s: %s\n", DRV_NAME, DRV_VERSION);

	ret = spi_register_driver(&w5500_driver);
	if(ret < 0)
	{
		printk("w5500 spi_register_driver failed\n");
		return ret;
	}
	else
		printk("w5500 spi register succeed\n");

	return 0;
}

static void __exit 
wiz_module_exit(void)
{
	spi_unregister_driver(&w5500_driver);
}

module_init(wiz_module_init);
module_exit(wiz_module_exit);

MODULE_AUTHOR("WIZnet");
MODULE_DESCRIPTION("Support for WIZnet w5500-based MACRAW Mode.");
MODULE_SUPPORTED_DEVICE("WIZnet W5500 Chip");
MODULE_LICENSE("GPL");

